@@ -3,12 +3,19 @@
  * by Stephen C. Hemminger
  *	linus!sch	or	sch@Mitre-Bedford
  *
- * [John Elliott, 18-7-1998] Updated to match the official LBR spec 
+ * [John Elliott, 2005-01-29] Merge patch from Steven N. Hirsch for 
+ * compilation on SuSE Linux 9.2
+ *      <jce@seasip.demon.co.uk>
+ *
+ * [John Elliott, 2003-01-11] Made all fopen()s use binary mode.
+ *      <jce@seasip.demon.co.uk>
+ *
+ * [John Elliott, 1998-07-18] Updated to match the official LBR spec 
  * (LUDEF5.DOC). This includes CRC checking (so archives made by lar 
  * aren't frowned at by CP/M archivers). 
  *      <jce@seasip.demon.co.uk>
  */
-#define VERSION "5.10"
+#define VERSION "5.1.2"
 /*
  *
  * Also added prototypes, and tried to shut 'gcc -Wall' up
@@ -25,6 +32,7 @@
  *	r - Reorganise library
  *  Other keys:
  *	v - Verbose
+ *	l - Produce ls -l listing
  *
  *  This program is public domain software, no warranty intended or
  *  implied.
@@ -58,11 +66,14 @@
  */
 
 #include <stdio.h>
+#include <assert.h>	/* [JCE] */
 #include <stdlib.h>	/* [JCE] */
 #include <unistd.h>	/* [JCE] */
 #include <sys/stat.h>	/* [JCE] */
 #include <time.h>	/* [JCE] */
 #include <ctype.h>
+#include <string.h>
+#include <errno.h>
 
 /* [JCE] CRC code */
 #include "crcsubs.h"
@@ -166,7 +177,13 @@ char  **argv;
     char   *aname;			/* name of library file */
     int	   (*function)(char *) = NULL;	/* function to do on library */
 /* set the function to be performed, but detect conflicts */
-#define setfunc(val)	if(function != NULL) conflict(); else function = val
+#define setfunc(val)	do { if(function != NULL) conflict(); \
+			     else function = val; \
+			} while(0) 
+
+/* 5.1.1: Based on email from C. B. Falconer: Check that the structure size
+ * is correct */
+    assert(DSIZE == 32);
 
     /* [JCE] init CRC code */
     CRC_Init(crcbuf);
@@ -211,6 +228,7 @@ char  **argv;
 	default: 
 	    help ();
     }
+    if (verbose && ls_l) conflict();  /* cbf */
 
     if(function == NULL) {
 	fprintf(stderr,"No function key letter specified\n");
@@ -224,7 +242,7 @@ char  **argv;
 /* print error message and exit */
 void 
 help (void ) {
-    fprintf (stderr, "Usage: %s {utepdr}[v] library [files] ...\n", cmdname);
+    fprintf (stderr, "Usage: %s {utepdr}[vl] library [files] ...\n", cmdname);
     fprintf (stderr, "Functions are:\n\tu - Update, add files to library\n");
     fprintf (stderr, "\tt - Table of contents\n");
     fprintf (stderr, "\te - Extract files from library\n");
@@ -263,10 +281,7 @@ void
 cant (name)
 char   *name;
 {
-    extern int  errno;
-    extern char *sys_errlist[];
-
-    fprintf (stderr, "%s: %s\n", name, sys_errlist[errno]);
+    fprintf (stderr, "%s: %s\n", name, strerror(errno));
     exit (1);
 }
 
@@ -296,8 +311,10 @@ char  **av;
 }
 
 
-int lelen(struct ludir *d)	/* [JCE] Calculate the precise length of an */
-{				/*       entry */
+int 
+lelen(d)			/* [JCE] Calculate the precise length of an */
+struct ludir *d;		/*       entry */
+{
     int i;
 
     i = 128 * wtoi(d->l_len);
@@ -316,16 +333,16 @@ char   *lib;
     int active = 0, unused = 0, deleted = 0;
     char *uname;
 
-    if ((lfd = fopen (lib, "r")) == NULL)
+    if ((lfd = fopen (lib, "rb")) == NULL)
 	cant (lib);
 
     getdir (lfd);
     total = wtoi(ldir[0].l_len);
     if(verbose) {
- 	printf("Name          Index Length   CRC       Date "
-               "created       Date updated \n");
+ 	printf("Name          Index Length   CRC         Dat"
+               "e created         Date updated \n");
         printf("============================================"
-               "==========================\n");
+               "==============================\n");
 	printf("Directory           %6d  %04x  ", total * 128, 
                                                   wtoi(ldir[0].l_crc));
         print_times(&ldir[0]);
@@ -338,6 +355,7 @@ char   *lib;
 		active++;
 		uname = getname(ldir[i].l_name, ldir[i].l_ext);
 		if (filarg (uname))
+                {
 		    if(verbose)
                     {
 			printf ("%-12s   %4d %6d  %04x  ", uname,
@@ -346,16 +364,15 @@ char   *lib;
                         print_times(&ldir[i]);
 			putchar('\n');
                     }
-		    else
-                    if (ls_l)
+		    else if (ls_l)
                     {
                         printf("-rw-rw-rw- 1 %8d %8d %8d ", 
                                getuid(), getgid(), lelen(&ldir[i]));
                         print_mtime(&ldir[i]);
                         printf(" %s\n", uname);
                     }
-                    else
-			printf ("%s\n", uname);
+                    else printf ("%s\n", uname);
+                }
 		total += wtoi(ldir[i].l_len);
 		break;
 	case UNUSED:
@@ -365,10 +382,10 @@ char   *lib;
 		deleted++;
 	}
     if(verbose) {
-	printf("-----------------------------------------"
+	printf("---------------------------------------------"
                "-----------------------------\n");
 	printf("Total bytes used    %6d\n", total * 128);
-	printf("\nLibrary %s has %d slots, %d deleted %d active, %d unused\n",
+	printf("\nLibrary %s has %d slots: %d deleted, %d active, %d unused, 1 directory.\n",
 		lib, nslots, deleted, active, unused);
     }
 
@@ -477,8 +494,9 @@ FILE *f;
 	else
 	    break;
     }
-
-    numsecs = nslots / SLOTS_SEC;
+/* 5.1.1: Based on email from C. B. Falconer: Add 1 slot for the first entry
+ * (the directory itself) */
+    numsecs = (nslots / SLOTS_SEC) + 1;
     nslots = numsecs * SLOTS_SEC;
 
     for (i = 0; i < nslots; i++)
@@ -591,7 +609,7 @@ bool	pflag;
     char   *unixname;
     word16 crc;		/* [JCE] */
 
-    if ((lfd = fopen (name, "r"))  == NULL)
+    if ((lfd = fopen (name, "rb"))  == NULL)
 	cant (name);
 
     ofd = pflag ? stdout : NULL;
@@ -605,7 +623,7 @@ bool	pflag;
 	    continue;
 	fprintf(stderr,"%s", unixname);
 	if (ofd != stdout)
-	    ofd = fopen (unixname, "w");
+	    ofd = fopen (unixname, "wb");
 	if (ofd == NULL) {
 	    fprintf (stderr, "  - can't create");
 	    errcnt++;
@@ -671,8 +689,8 @@ char   *name;
     FILE *lfd;
     register int    i;
 
-    if ((lfd = fopen (name, "r+")) == NULL) {
-	if ((lfd = fopen (name, "w+")) == NULL)
+    if ((lfd = fopen (name, "r+b")) == NULL) {
+	if ((lfd = fopen (name, "w+b")) == NULL)
 	    cant (name);
 	initdir (lfd);
     }
@@ -708,7 +726,7 @@ FILE *lfd;
         return;
     }
 
-    if ((ifd = fopen (name, "r")) == NULL) {
+    if ((ifd = fopen (name, "rb")) == NULL) {
 	fprintf (stderr, "%s: can't find to add\n",name);
 	errcnt++;
 	return;
@@ -782,7 +800,7 @@ char   *lname;
     FILE *f;
     register int    i;
 
-    if ((f = fopen (lname, "r+")) == NULL)
+    if ((f = fopen (lname, "r+b")) == NULL)
 	cant (lname);
 
     if (nfiles <= 0)
@@ -816,10 +834,10 @@ char  *name;
 
     VOID sprintf(tmpname,"%-10.10s.TMP", name);
 
-    if( (olib = fopen(name,"r")) == NULL)
+    if( (olib = fopen(name,"rb")) == NULL)
 	cant(name);
 
-    if( (nlib = fopen(tmpname, "w")) == NULL)
+    if( (nlib = fopen(tmpname, "wb")) == NULL)
 	cant(tmpname);
 
     getdir(olib);
@@ -871,6 +889,13 @@ FILE *of, *nf;
     new->l_stat = ACTIVE;
     copymem(new->l_name, old->l_name, 8);
     copymem(new->l_ext, old->l_ext, 3);
+/* [JCE v5.1.1] Copy these across as well when reorganising */
+    new->l_crc   = old->l_crc;
+    new->l_datec = old->l_datec;
+    new->l_dateu = old->l_dateu;
+    new->l_timec = old->l_timec;
+    new->l_timeu = old->l_timeu;
+    new->l_pad   = old->l_pad;
     VOID fseek(of, (long) wtoi(old->l_off)*SECTOR, 0);
     VOID fseek(nf, 0L, 2);
     secoffs = ftell(nf) / SECTOR;
@@ -1011,17 +1036,24 @@ int days;
 {
 	if (ls_l)
         {
+/* Keep the ls -l format using 2-digit years; mc may require it */
+/*		if (!days) printf("01-01-1970 00:00");
+		else       printf("%02d-%02d-%04d %02d:%02d",  
+                          ptm->tm_mon+1, ptm->tm_mday, (ptm->tm_year + 1900),
+                          ptm->tm_hour, ptm->tm_min);
+*/
 		if (!days) printf("01-01-70 00:00");
 		else       printf("%02d-%02d-%02d %02d:%02d",  
                           ptm->tm_mon+1, ptm->tm_mday, (ptm->tm_year % 100),
                           ptm->tm_hour, ptm->tm_min);
+
 		return;
         }
 	
 
 	if (!days) printf("%19s","");
-	else	printf("%02d/%02d/%02d %02d:%02d:%02d  ", 
-               ptm->tm_mday, ptm->tm_mon+1, (ptm->tm_year % 100),       
+	else	printf("%02d/%02d/%04d %02d:%02d:%02d  ", 
+               ptm->tm_mday, ptm->tm_mon+1, (ptm->tm_year + 1900),       
                ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
 }
 
